use serde::{de::DeserializeOwned, Deserialize, Serialize};

use crate::{col, get_mongo, id_of};

use super::valid::Validate;

/// A `Reference` to a `Model`
#[derive(Debug, Clone, Default, Serialize, Deserialize, PartialEq, Eq)]
pub struct Reference(String);

impl Reference {
    /// Create a new reference from String
    pub async fn new_raw(reference: &str) -> Option<Self> {
        let r = Self(reference.to_string());
        if r.validate().await {
            Some(r)
        } else {
            None
        }
    }

    /// Create a new reference
    pub async fn new(model: &str, id: &str) -> Option<Self> {
        Self::new_raw(&format!("{model}::{id}")).await
    }

    /// Checks if a reference is part of a model collection.
    #[must_use]
    pub fn is_of_collection(&self, col: &str) -> bool {
        self.0.starts_with(col)
    }

    /// Get the raw `Document` behind the `Reference` from the database.
    pub async fn get_document(&self) -> Option<mongodb::bson::Document> {
        let (col, id) = self.0.split_once("::")?;
        let db = get_mongo!();
        let col = col!(db, col);
        col.find_one(id_of!(id), None).await.ok()?
    }

    /// Get the `Model` behind the `Reference` from the database.
    pub async fn get<T: DeserializeOwned>(&self) -> T {
        mongodb::bson::from_document(self.get_document().await.unwrap()).unwrap()
    }
}

impl Validate for Reference {
    async fn validate(&self) -> bool {
        // cheap
        //self.0.split_once("::").is_some()

        // right
        self.get_document().await.is_some()
    }
}

/// A trait that allows you to get a reference to a `Model`
pub trait Referencable {
    /// The name of this model's collection
    fn collection_name() -> &'static str;
    /// The id of this model
    fn id(&self) -> &str;

    /// Get a reference to the object
    fn reference(&self) -> Reference {
        Reference(format!("{}::{}", Self::collection_name(), self.id()))
    }
}
